/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** STM32F1 example
 *  @file main.c
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2016
 */

/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdio.h> // standard I/O facilities
#include <stdlib.h> // standard utilities
#include <unistd.h> // standard streams
#include <string.h> // string utilities
#include <math.h> // mathematical utilities

/* STM32 (including CM3) libraries */
#include <libopencmsis/core_cm3.h> // Cortex M3 utilities
#include <libopencm3/cm3/scb.h> // vector table definition
#include <libopencm3/cm3/nvic.h> // interrupt utilities
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/exti.h> // external interrupt utilities
#include <libopencm3/stm32/rtc.h> // real time clock utilities
#include <libopencm3/stm32/iwdg.h> // independent watchdog utilities
#include <libopencm3/stm32/dbgmcu.h> // debug utilities
#include <libopencm3/stm32/flash.h> // flash utilities

/* own libraries */
#include "global.h" // board definitions
//#include "usart.h" // USART utilities
#include "usb_cdcacm.h" // USB CDC ACM utilities
#include "sensor_pzem.h" // PZEM electricity meter utilities
#include "sensor_sdm120.h" // SDM120 electricity meter utilities
#include "radio_esp8266.h" // ESP8266 WiFi SoC utilities

#define WATCHDOG_PERIOD 10000 /**< watchdog period in ms */

/** @defgroup main_flags flag set in interrupts to be processed in main task
 *  @{
 */
volatile bool rtc_internal_tick_flag = false; /**< flag set when internal RTC ticked */
/** @} */

int _write(int file, char *ptr, int len)
{
	int i; // how much data has been sent
	static char newline = 0; // what newline has been sent

	if (file == STDOUT_FILENO || file == STDERR_FILENO) {
		for (i = 0; i < len; i++) {
			if (ptr[i] == '\r' || ptr[i] == '\n') { // send CR+LF newline for most carriage return and line feed combination
				if (newline==0 || (newline==ptr[i])) { // newline has already been detected
					//usart_putchar_nonblocking('\r'); // send newline over USART
					//usart_putchar_nonblocking('\n'); // send newline over USART
					cdcacm_putchar('\r'); // send newline over USB
					cdcacm_putchar('\n'); // send newline over USB
					newline = ptr[i]; // remember the newline
				}
				if (ptr[i] == '\n') { // line feed are always considered to end a line (the LF+CR combination is not supported to better support the others)
					newline = 0; // clear new line
				}
			} else { // non-newline character
				//usart_putchar_nonblocking(ptr[i]); // send byte over USART
				cdcacm_putchar(ptr[i]); // send byte over USB
				newline = 0; // clear new line
			}
		}
		return i;
	}
	return -1;
}

/** user input command */
static char command[32] = {0};
/** user input command index */
uint8_t command_i = 0;

/** process user command
 *  @param[in] str user command string (\0 ended)
 */
static void process_command(char* str)
{
	// split command
	const char* delimiter = " ";
	char* word = strtok(str,delimiter);
	if (!word) {
		goto error;
	}
	// parse command
	if (0==strcmp(word,"help")) {
		printf("available commands:\n");
		printf("led [on|off|toggle]\n");
		printf("time [HH:MM:SS]\n");
	} else if (0==strcmp(word,"led")) {
		word = strtok(NULL,delimiter);
		if (!word) {	
			goto error;
		} else if (0==strcmp(word,"on")) {
			led_on(); // switch LED on
			printf("LED switched on\n"); // notify user		
		} else if (0==strcmp(word,"off")) {
			led_off(); // switch LED off
			printf("LED switched off\n"); // notify user
		} else if (0==strcmp(word,"toggle")) {
			led_toggle(); // toggle LED
			printf("LED toggled\n"); // notify user
		} else {
			goto error;
		}
	} else if (0==strcmp(word,"time")) {
		word = strtok(NULL,delimiter);
		if (!word) {
			printf("current time: %02lu:%02lu:%02lu\n", rtc_get_counter_val()/(60*60), (rtc_get_counter_val()%(60*60))/60, (rtc_get_counter_val()%60)); // get and print time from internal RTC
		} else if (strlen(word)!=8 || word[0]<'0' || word[0]>'2' || word[1]<'0' || word[1]>'9' || word[3]<'0' || word[3]>'5' || word[4]<'0' || word[4]>'9' || word[6]<'0' || word[6]>'5' || word[7]<'0' || word[7]>'9') { // time format is incorrect
				goto error;
		} else {
			rtc_set_counter_val(((word[0]-'0')*10+(word[1]-'0')*1)*(60*60)+((word[3]-'0')*10+(word[4]-'0')*1)*60+((word[6]-'0')*10+(word[7]-'0')*1)); // set time in internal RTC counter
			printf("time set\n");
		}
	} else {
		goto error;
	}

	return; // command successfully processed
error:
	printf("command not recognized. enter help to list commands\n");
	return;
}

/** send HTTP data
 *  @warning blocking until a response has been received
 *  @param[in] data data to be send
 *  @param[in] length number of bytes to be sent, set to 0 to use the string length
 *  @return if data has been sent
 */
static bool http_send(uint8_t* data, size_t length)
{
	if (length==0) {
		radio_esp8266_send(data, strlen((char*)data)); // send string data
	} else {
		radio_esp8266_send(data, length); // send raw data
	}
	while (!radio_esp8266_activity) { // wait until response has been received
		__WFI(); // wait until something happens
	}
	if (!radio_esp8266_success) {
		fprintf(stderr,"could not send data\n");
		return false;
	}
	return true;
}

/** end HTTP connection
 *  @warning blocking until a response has been received
 *  @return if connection has been closed
 */
static bool http_end(void)
{
	radio_esp8266_close(); // close connection
	while (!radio_esp8266_activity) { // wait until response has been received
		__WFI(); // wait until something happens
	}
	return radio_esp8266_success;
}

/** open HTTP connection and send POST header
 *  @warning blocking until a response has been received
 *  @param[in] host host name or IP of HTTP server to connect to
 *  @param[in] port port number of HTTP server to connect to
 *  @param[in] length number of bytes to POST
 *  @return if HTTP POST succeeded
 */
static bool http_post_header(char* host, uint16_t port, size_t length)
{
	char http_line[256] = {0}; // generated lines
	radio_esp8266_tcp_open(host, port); // open connection
	while (!radio_esp8266_activity) { // wait until response has been received
		__WFI(); // wait until something happens
	}
	if (!radio_esp8266_success) {
		fprintf(stderr,"TCP connection failed\n");
		return false;
	}
	if (!http_send((uint8_t*)"POST /write?db=test HTTP/1.1\r\n", 0)) { // send data
		return false;
	}
	if (snprintf(http_line, LENGTH(http_line), "Content-Length: %u\r\n", length)<0) { // set content length (for measurements)
		fprintf(stderr,"could not create line\n");
		return false;
	}
	if (!http_send((uint8_t*)http_line, 0)) { // send data
		return false;
	}
    if (!http_send((uint8_t*)"Host: influx\r\n", 0)) { // send data
		 return false;
	}
    if (!http_send((uint8_t*)"\r\n", 0)) { // send data
		 return false;
	}
	return true;
}

/** program entry point
 *  this is the firmware function started by the micro-controller
 */
void main(void);
void main(void)
{	
	rcc_clock_setup_in_hse_8mhz_out_72mhz(); // use 8 MHz high speed external clock to generate 72 MHz internal clock

#if DEBUG
	DBGMCU_CR |= DBGMCU_CR_IWDG_STOP; // stop independent watchdog counter when code is halted
	DBGMCU_CR |= DBGMCU_CR_WWDG_STOP; // stop window watchdog counter when code is halted
	DBGMCU_CR |= DBGMCU_CR_STANDBY; // allow debug also in standby mode (keep digital part and clock powered)
	DBGMCU_CR |= DBGMCU_CR_STOP; // allow debug also in stop mode (keep clock powered)
	DBGMCU_CR |= DBGMCU_CR_SLEEP; // allow debug also in sleep mode (keep clock powered)
#endif

	// setup board
	board_setup();
	
	// setup USART and USB for user communication
    //usart_setup(); // setup USART (for printing)
	cdcacm_setup(); // setup USB CDC ACM (for printing)
	setbuf(stdout, NULL); // set standard out buffer to NULL to immediately print
	setbuf(stderr, NULL); // set standard error buffer to NULL to immediately print

	// minimal setup ready	
	printf("welcome to the STM32F1 CuVoodoo example code\n"); // print welcome message

	// setup RTC
	printf("setup internal RTC: ");
	rtc_auto_awake(RCC_LSE, 32768-1); // ensure internal RTC is on, uses the 32.678 kHz LSE, and the prescale is set to our tick speed, else update backup registers accordingly (power off the micro-controller for the change to take effect)
	rtc_interrupt_enable(RTC_SEC); // enable RTC interrupt on "seconds"
	nvic_enable_irq(NVIC_RTC_IRQ); // allow the RTC to interrupt
	printf("OK\n");

	uint32_t ticks_time = rtc_get_counter_val(); // get time from internal RTC (since first start/power up)
	printf("uptime: %02lu:%02lu:%02lu\n", ticks_time/(60*60), (ticks_time%(60*60))/60, (ticks_time%60)); // display time

	// setup PZEM electricity meter
	printf("setup PZEM-004 electricity meter: ");
	sensor_pzem_setup(); // setup PZEM electricity meter
	printf("OK\n");

	// setup SDM120 electricity meter
	printf("setup SDM120 electricity meter: ");
	sensor_sdm120_setup(9600); // setup SDM120 electricity meter (get baud rate by scrolling through the menu on the device)
	printf("OK\n");

	//setup ESP8266 WiFi SoC
	printf("setup ESP8266 WiFi SoC: ");
	radio_esp8266_setup();
	printf("OK\n");

#if !(DEBUG)
	//setup watchdog to reset in case we get stuck (i.e. when an error occurred)
	printf("setup watchdog (%.2fs): ",WATCHDOG_PERIOD/1000.0);
	iwdg_set_period_ms(WATCHDOG_PERIOD); // set independent watchdog period
	iwdg_start(); // start independent watchdog
	printf("OK\n");
	if (FLASH_OBR&FLASH_OBR_OPTERR) {
		printf("option bytes not set in flash: software wachtdog used (not started at reset)\n");
	} else if (FLASH_OBR&FLASH_OBR_WDG_SW) {
		printf("software wachtdog used (not started at reset)\n");
	} else {
		printf("hardware wachtdog used (started at reset)\n");
	}
#endif

	// send HTTP POST request
	printf("making HTTP request: ");
	char line[256] = {0}; // measurement line to send
	if (snprintf(line, LENGTH(line), "cpu_load_short,host=server01,region=us-west value=0.64 1434055562000000000")<0) {
		fprintf(stderr,"could not create line\n");
	} else if (!http_post_header("192.168.42.3", 8086, strlen(line))) { // send header
		fprintf(stderr,"could not sent HTTP POST header\n");
	} else if (!http_send((uint8_t*)line, 0)) { // send data
		fprintf(stderr,"could not send measurement\n");
	} else {
		http_end(); // end HTTP request (don't care about the result)
		printf("OK\n");
	}
	
	// main loop
	printf("command input: ready\n");
	bool action = false; // if an action has been performed don't go to sleep
	button_flag = false; // reset button flag
	char c = ' '; // to store received character
	bool char_flag = false; // a new character has been received

	// start getting all PZEM-004T measurements from all meters
	struct sensor_pzem_measurement_t pzem_measurements[2][SENSOR_PZEM_MAX]; // PZEM-004T measurements (2 meters, all measurements)
	uint8_t pzem_id = 0; // PZEM-004T meter index (add to prefix)
	sensor_pzem_measurement_request(0xc0a80100+pzem_id, 0); // request first measurement
	(void)pzem_measurements[pzem_id][0];

	// start getting all SDM120 measurements from all meters
	float sdm120_measurements[2][SENSOR_SDM120_MEASUREMENT_MAX]; // SDM120 measurements (2 meters, all measurements)
	uint8_t sdm120_meter = 0; // SDM120 meter index (add to 1 to get ID)
	uint8_t sdm120_measurement = 0; // SDM120 measurement index
	sensor_sdm120_measurement_request(1+sdm120_meter, sdm120_measurement); // request first measurement
	(void)sdm120_measurements[sdm120_meter][sdm120_measurement];

	while (true) { // infinite loop
		iwdg_reset(); // kick the dog
/*
		while (usart_received) { // data received over UART
			action = true; // action has been performed
			led_toggle(); // toggle LED
			c = usart_getchar(); // store receive character
			char_flag = true; // notify character has been received
		}
*/
		while (cdcacm_received) { // data received over USB
			action = true; // action has been performed
			led_toggle(); // toggle LED
			c = cdcacm_getchar(); // store receive character
			char_flag = true; // notify character has been received
		}
		while (char_flag) { // user data received
			char_flag = false; // reset flag
			action = true; // action has been performed
			printf("%c",c); // echo receive character
			if (c=='\r' || c=='\n') { // end of command received
				if (command_i>0) { // there is a command to process
					command[command_i] = 0; // end string
					command_i = 0; // prepare for next command
					process_command(command); // process user command
				}
			} else { // user command input
				command[command_i] = c; // save command input
				if (command_i<LENGTH(command)-2) { // verify if there is place to save next character
					command_i++; // save next character
				}
			}
		}
		while (sensor_pzem_measurement_received) { // measurement from electricity meter received
			struct sensor_pzem_measurement_t measurement = sensor_pzem_measurement_decode(); // decode measurement
			if (measurement.type>=SENSOR_PZEM_MAX) {
				fprintf(stderr,"unknown measurement type: %u\n", measurement.type);
				while (true); // unhandled error
			}
			pzem_measurements[pzem_id][measurement.type] = measurement; // save measurement (the type matches the index)
			if (measurement.valid) { // only show valid measurement
				switch (measurement.type) {
				case SENSOR_PZEM_VOLTAGE:
					printf("PZEM-004T meter %u voltage: %.01f V\n", pzem_id, measurement.value.voltage); // display measurement
					break;
				case SENSOR_PZEM_CURRENT:
					printf("PZEM-004T meter %u current: %.02f A\n", pzem_id, measurement.value.current);
					break;
				case SENSOR_PZEM_POWER:
					printf("PZEM-004T meter %u power: %.00f W\n", pzem_id, measurement.value.power);
					break;
				case SENSOR_PZEM_ENERGY:
					printf("PZEM-004T meter %u energy: %lu Wh\n", pzem_id, measurement.value.energy);
					break;
/* not used for this application
				case SENSOR_PZEM_ADDRESS:
					printf("PZEM-004T meter %u address set\n", pzem_id);
					break;
				case SENSOR_PZEM_ALARM:
					printf("PZEM-004T meter %u alarm threshold set\n", pzem_id);
					break;
*/
				default:
					break;
				}
				if (measurement.type<SENSOR_PZEM_MAX-1) { // not all measurement types requested
					 sensor_pzem_measurement_request(0xc0a80100+pzem_id, measurement.type+1); // request next measurement
				} else { // all measurement types requested
					pzem_id++; // got to next meter
					if (pzem_id<LENGTH(pzem_measurements)) { // ensure next meter exists
						sensor_pzem_measurement_request(0xc0a80100+pzem_id, 0); // request measurement for next meter
					}
				}
			} else { // measurement not valid
				sensor_pzem_measurement_request(0xc0a80100+pzem_id, measurement.type+1); // request same measurement
			}
		}
		while (sensor_sdm120_measurement_received) { // measurement from electricity meter received
			float measurement = sensor_sdm120_measurement_decode(); // decode measurement
			if (isnan(measurement)) {
				printf("error in SDM120 response\n");
				sensor_sdm120_measurement_request(1+sdm120_meter, sdm120_measurement); // request same measurement
			} else if (isinf(measurement)) {
				printf("error SDM120 message received\n");
				while (true); // unhandled error
			} else {
				sdm120_measurements[sdm120_meter][sdm120_measurement] = measurement; // save measurement
				printf("SDM120 meter %u measurement %u: %.01f\n", sdm120_meter, sdm120_measurement, measurement);
				if (sdm120_measurement<SENSOR_SDM120_MEASUREMENT_MAX-1) { // not all measurement type required
					sdm120_measurement++; // go to next measurement
					sensor_sdm120_measurement_request(1+sdm120_meter, sdm120_measurement); // request next measurement
				} else { // all measurement types requested
					sdm120_meter++; // got to next meter (sending to none existing meter will just end in void)
					sdm120_measurement = 0; // start requesting all measurement
					if (sdm120_meter<LENGTH(sdm120_measurements)) { // ensure next meter exists
						sensor_sdm120_measurement_request(1+sdm120_meter, sdm120_measurement); // request measurement for next meter
					}
				}	
			}
		}
		while (button_flag) { // user pressed button
			action = true; // action has been performed
			printf("button pressed\n");
			led_toggle(); // toggle LED
			for (uint32_t i=0; i<1000000; i++) { // wait a bit to remove noise and double trigger
				__asm__("nop");
			}
			button_flag = false; // reset flag
		}
		while (rtc_internal_tick_flag) { // the internal RTC ticked
			rtc_internal_tick_flag = false; // reset flag
			led_toggle(); // toggle LED (good to indicate if main function is stuck)
			ticks_time = rtc_get_counter_val(); // copy time from internal RTC for processing
			action = true; // action has been performed
			if ((ticks_time%(60))==0) { // one minute passed
				printf("uptime: %02lu:%02lu:%02lu\n", ticks_time/(60*60), (ticks_time%(60*60))/60, (ticks_time%60)); // display external time
			}
		}
		if (action) { // go to sleep if nothing had to be done, else recheck for activity
			action = false;
		} else {
			__WFI(); // go to sleep
		}
	}
}


/** @brief interrupt service routine called when tick passed on RTC */
void rtc_isr(void)
{
	rtc_clear_flag(RTC_SEC); // clear flag
	rtc_internal_tick_flag = true; // notify to show new time
}
